use std::fmt::{Debug, Formatter};
use std::ops::{Deref, DerefMut};

use bevy::math::UVec2;
use bevy::prelude::*;
use fastrand::Rng;

use crate::world::adjacency::{
	get_floor_sprite_offset, get_wall_sprite_offset, BOTTOM, LEFT, NONE, RIGHT, TOP,
};
use crate::world::generators::MapGenerator;

pub const WORLD_TILE_SIZE: f32 = 16.0;

/// Marker component for any spawned tile entity
#[derive(Default, Debug, Copy, Clone, Eq, PartialEq, Component)]
pub struct Tile;

/// Track the location of an entity within a grid
#[derive(Default, Debug, Copy, Clone, Eq, PartialEq, Component)]
#[repr(transparent)]
pub struct GridPosition(pub UVec2);
impl Deref for GridPosition {
	type Target = UVec2;
	fn deref(&self) -> &Self::Target {
		&self.0
	}
}
impl DerefMut for GridPosition {
	fn deref_mut(&mut self) -> &mut Self::Target {
		&mut self.0
	}
}

/// Take an entity's position on a grid, and sync it up to the transform used to render
/// the sprite
pub fn sync_grid_to_transform(mut query: Query<(&GridPosition, &mut Transform)>) {
	for (position, mut transform) in &mut query {
		transform.translation = ((position.as_vec2() * WORLD_TILE_SIZE) + (WORLD_TILE_SIZE / 2.0))
			.extend(transform.translation.z);
	}
}

/// Store accessibility of a tile and the sprite group that the tile belongs to.
/// Used for pathfinding, and for selecting which sprites should be considered as
/// a group
#[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
pub struct MapTile {
	pub can_walk: bool,
	pub can_fly: bool,
	pub can_see: bool,
	pub tile_group: usize,
}

impl MapTile {
	/// Create a tile that can be passed by all entities
	pub fn new_floor(tile_group: usize) -> MapTile {
		MapTile {
			can_walk: true,
			can_fly: true,
			can_see: true,
			tile_group,
		}
	}

	/// Create a tile that can be passed by no entities
	pub fn new_wall(tile_group: usize) -> MapTile {
		MapTile {
			can_walk: false,
			can_fly: false,
			can_see: false,
			tile_group,
		}
	}
	/// Create a tile that can be passed by flying entities
	/// and seen through by any entity
	pub fn new_obstacle(tile_group: usize) -> MapTile {
		MapTile {
			can_walk: false,
			can_fly: true,
			can_see: true,
			tile_group,
		}
	}
}

#[derive(Clone, Eq, PartialEq)]
pub enum MapLayerType {
	Wall,
	Floor,
	None,
}

impl MapLayerType {
	pub fn transform_adjacency(&self, value: usize, adjacency: u8) -> usize {
		match self {
			MapLayerType::Wall => value + get_wall_sprite_offset(adjacency),
			MapLayerType::Floor => value + get_floor_sprite_offset(adjacency),
			MapLayerType::None => value,
		}
	}
}

#[derive(Clone, Eq, PartialEq)]
pub struct MapLayer {
	pub width: usize,
	pub height: usize,
	pub offset_horizontal: usize,
	pub offset_vertical: usize,
	pub tiles: Vec<Option<MapTile>>,
	pub layer_type: MapLayerType,
}

impl MapLayer {
	pub fn from_sized_list(
		width: usize,
		height: usize,
		tiles: Vec<Option<MapTile>>,
		layer_type: MapLayerType,
	) -> Self {
		MapLayer {
			width,
			height,
			offset_horizontal: 0,
			offset_vertical: 0,
			tiles,
			layer_type,
		}
	}

	pub fn get_tile_adjacency(&self, x: usize, y: usize) -> u8 {
		let indexer = Indexer::new(self.width, self.height);
		let mut adjacency = NONE;

		if x > 0 && self.tiles[indexer.index(x - 1, y)].is_some() {
			adjacency = adjacency | LEFT;
		}

		if x < self.width - 1 && self.tiles[indexer.index(x + 1, y)].is_some() {
			adjacency = adjacency | RIGHT;
		}

		if y > 0 && self.tiles[indexer.index(x, y - 1)].is_some() {
			adjacency = adjacency | BOTTOM;
		}

		if y < self.height - 1 && self.tiles[indexer.index(x, y + 1)].is_some() {
			adjacency = adjacency | TOP;
		}

		adjacency
	}
}

impl Debug for MapLayer {
	fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
		writeln!(f, "---\nMap Layer: {}x{}", self.width, self.height)?;
		let idx = Indexer::new(self.width, self.height);
		for y in (0..self.height).rev() {
			for x in 0..self.width {
				write!(
					f,
					"{}",
					self.tiles[idx.index(x, y)]
						.map(|t| t.tile_group.to_string())
						.unwrap_or(" ".to_string())
				)?;
			}
			write!(f, "\n")?;
		}

		writeln!(f, "---")
	}
}

#[derive(Debug, Clone, Eq, PartialEq, Component)]
pub struct LevelMap {
	pub width: usize,
	pub height: usize,
	pub layers: Vec<MapLayer>,
	pub spawn: UVec2,
}

impl LevelMap {
	pub fn new(width: usize, height: usize, spawn: UVec2, layers: Vec<MapLayer>) -> Self {
		LevelMap {
			width,
			height,
			layers,
			spawn,
		}
	}
}

#[derive(Bundle)]
pub struct LevelMapBundle {
	level_map: LevelMap,
	#[bundle]
	transform: TransformBundle,
	#[bundle]
	visibility: VisibilityBundle,
}

impl LevelMapBundle {
	pub fn new(level_map: LevelMap) -> Self {
		Self {
			level_map,
			transform: TransformBundle::default(),
			visibility: VisibilityBundle::default(),
		}
	}

	pub fn generate<Gen: MapGenerator>(width: usize, height: usize) -> Self {
		Self::generate_with_rng::<Gen>(&Rng::new(), width, height)
	}
	pub fn generate_with_seed<Gen: MapGenerator>(seed: u64, width: usize, height: usize) -> Self {
		Self::generate_with_rng::<Gen>(&Rng::with_seed(seed), width, height)
	}
	pub fn generate_with_rng<Gen: MapGenerator>(rng: &Rng, width: usize, height: usize) -> Self {
		let indexer = Indexer::new(width, height);
		LevelMapBundle::new(Gen::generate(&indexer, rng))
	}
}

#[derive(Debug, Copy, Clone)]
pub struct Indexer {
	width: usize,
	height: usize,
}

impl Indexer {
	pub fn new(width: usize, height: usize) -> Self {
		Indexer { width, height }
	}
	pub fn index(&self, x: usize, y: usize) -> usize {
		(y * self.width) + x
	}
	pub fn reverse(&self, index: usize) -> (usize, usize) {
		(index % self.width, index / self.width)
	}
	pub fn index_within(&self, idx: usize) -> bool {
		let (x, y) = self.reverse(idx);
		x >= 0 && x < self.width && y >= 0 && y < self.height
	}
	pub fn width(&self) -> usize {
		self.width
	}
	pub fn height(&self) -> usize {
		self.height
	}
}
